import re

import logging

from urllib.parse import urlparse, urljoin
from html.parser import HTMLParser

from django.conf import settings


LOG = logging.getLogger('tests')


class LinkExtractor:

    def extract(self, content):
        raise NotImplementedError


class DummyLinkExtractor(LinkExtractor):

    def extract(self, content):
        return []


class HTMLLinkExtractor(LinkExtractor):

    link_attr_names = ('href', 'src')

    def extract(self, content):
        link_attr_names = self.link_attr_names

        class SaxLinkExtractor(HTMLParser):
            links = set()

            def handle_starttag(self, tag, attrs):
                self.links.update(
                    v for k, v in attrs if k in link_attr_names
                )

        parser = SaxLinkExtractor()
        parser.feed(content)
        parser.close()

        return parser.links


class URLWithReferer:

    def __init__(self, url, referer=None):
        self.url = url
        self.referer = referer

    def __eq__(self, other):
        if isinstance(other, URLWithReferer):
            return self.url == other.url
        if isinstance(object, str):
            return self.url == other
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.url

    def __hash__(self):
        return hash(self.url)

    def __unicode__(self):
        return self.url


class Crawler:

    def __init__(self, client, base_urls=None, exclude_urls=None, pre_request=None, post_response=None,
                 extra_link_extractors=None):
        self.client = client
        self.urls = set()
        if base_urls:
            for url in base_urls:
                self.urls.add(URLWithReferer(url))

        self.crawled_urls = set()
        self.exclude_urls = set(exclude_urls or ())
        self.pre_request = pre_request
        self.post_response = post_response
        self.link_extractors = {
            'text/html': HTMLLinkExtractor(),
            'default': DummyLinkExtractor(),
        }
        if extra_link_extractors:
            self.link_extractors.update(extra_link_extractors)
        self.excluded_urls_regexs = [re.compile(r'^{}$'.format(url)) for url in self.exclude_urls]

    def run(self):
        while self.urls:
            self._call_request(self.urls.pop())

    def _parse_urls(self, url, resp):
        returned_urls = []
        content_type = resp['Content-Type'].split(';')[0]
        if content_type in self.link_extractors:
            content = resp.content
            if 'utf-8' in resp['Content-Type']:
                content = content.decode('utf-8')

            for link in self.link_extractors.get(content_type, self.link_extractors['default']).extract(content):
                parsed_href = urlparse(link)

                if not parsed_href.path:
                    continue

                if parsed_href.scheme and not parsed_href.netloc.startswith('testserver'):
                    LOG.debug('Skipping external link: %s', link)
                    continue

                if ('django.contrib.staticfiles' in settings.INSTALLED_APPS
                        and parsed_href.path.startswith(settings.STATIC_URL)):
                    LOG.debug('Skipping static file %s', parsed_href.path)
                elif parsed_href.path.startswith(settings.MEDIA_URL):
                    LOG.debug('Skipping media file %s', parsed_href.path)
                elif parsed_href.path.startswith('/'):
                    returned_urls.append(link)
                else:
                    # We'll use urlparse's urljoin since that handles things like <a href="../foo">
                    returned_urls.append(urljoin(url, link))

        return returned_urls

    def _pre_request(self, url, referer, headers):
        if self.pre_request:
            self.pre_request(url, referer, headers)

        return url, headers

    def _post_response(self, url, referer, resp=None, exception=None):
        if self.post_response:
            self.post_response(url, referer, resp, exception)

    def _call_request(self, url_with_referer):
        url, referer = url_with_referer.url, url_with_referer.referer
        resp = None
        try:
            url, headers = self._pre_request(url, referer, {})
            resp = self.client.get(url, follow=True, **headers)
            self.crawled_urls.add(url)
            if resp.redirect_chain:
                self.crawled_urls.update((redirect_url for redirect_url, _ in resp.redirect_chain))
            for parsed_url in self._parse_urls(url, resp):
                if (parsed_url not in self.crawled_urls and parsed_url not in self.urls and
                        not any(excluded_urls_regex.match(parsed_url)
                                for excluded_urls_regex in self.excluded_urls_regexs)):
                    self.urls.add(URLWithReferer(parsed_url, url))
            self._post_response(url, referer, resp)
        except Exception as e:
            LOG.exception('%s had unhandled exception: %s', url, e)
            self._post_response(url, referer, resp, e)



import json

from django.utils.encoding import force_str

from .trivials import assert_true, assert_equal, assert_in, fail
from .http import assert_http_ok, assert_http_created


def assert_valid_JSON(data, msg='Json is not valid'):
    """
    Given the provided ``data`` as a string, ensures that it is valid JSON &
    can be loaded properly.
    """
    try:
        json.loads(force_str(data))
    except Exception:
        fail(msg)


def assert_valid_JSON_response(resp, msg=None):
    """
    Given a ``HttpResponse`` coming back from using the ``client``, assert that
    you get back:

    * An HTTP 200
    * The correct content-type (``application/json``)
    * The content is valid JSON
    """
    assert_http_ok(resp, msg)
    assert_true(resp['Content-Type'].startswith('application/json'), msg)
    assert_valid_JSON(resp.content, msg)


def assert_valid_JSON_created_response(resp, msg=None):
    """
    Given a ``HttpResponse`` coming back from using the ``client``, assert that
    you get back:

    * An HTTP 201
    * The correct content-type (``application/json``)
    * The content is valid JSON
    """
    assert_http_created(resp, msg)
    assert_true(resp['Content-Type'].startswith('application/json'), msg)
    assert_valid_JSON(resp.content, msg)


def assert_keys(data, expected, msg=None):
    """
    This method ensures that the keys of the ``data`` match up to the keys of
    ``expected``.

    It covers the (extremely) common case where you want to make sure the keys of
    a response match up to what is expected. This is typically less fragile than
    testing the full structure, which can be prone to data changes.
    """
    assert_equal(set(data.keys()), set(expected), msg)



import types
import re
from collections.abc import Iterable, Sequence,Mapping
from functools import wraps

from inspect import isclass, isfunction, ismethod, getfullargspec, signature

import responses
from django.db import transaction
from django.db.models import Model
from django.db.models.fields import DateField, DateTimeField
from django.utils.functional import cached_property


def is_iterable(data):
    return isinstance(data, Iterable) and not isinstance(data, str)


def refresh_model_object(obj):
    obj.refresh_from_db()
    for key, value in obj.__class__.__dict__.items():
        if isinstance(value, cached_property):
            obj.__dict__.pop(key, None)


def refresh_model_objects(*data):
    [refresh_model_object(obj) for obj in data if isinstance(obj, Model) and obj.pk]


def login(function=None, users_generator='get_user', **users_kwargs):
    """Login decorator usage: @login(user_data)"""

    def _login(function):
        def _decorator(self, *args, **kwargs):
            if isinstance(users_generator, str):
                users = getattr(self, users_generator)(**users_kwargs)
            else:
                users = users_generator(self, **users_kwargs)

            if not isinstance(users, (list, tuple)):
                users = (users,)

            for user in users:
                self.login(user)
                function(self, *args, **kwargs)
                self.logout()
        return wraps(function)(_decorator)

    if function:
        return _login(function)
    return _login


def login_all(cls=None, **user_kwargs):
    """Login decorator for all methods inside test usage: @login(user_data)"""

    def _login_all(cls):
        for attr, val in cls.__dict__.items():
            if callable(val) and attr.startswith("test_"):
                setattr(cls, attr, login(**user_kwargs)(val))
        return cls

    if cls:
        return _login_all(cls)
    return _login_all


def fill_data_with_source(data, named_data):
    if not named_data:
        return data
    if isinstance(data, dict):
        return {fill_data_with_source(k, named_data): fill_data_with_source(v, named_data) for k, v in data.items()}
    elif isinstance(data, tuple) and hasattr(data, '_fields'):
        # named tuples require special handling
        return type(data)(
            *(fill_data_with_source(v, named_data) for v in data)
        )
    elif isinstance(data, (list, tuple, set)):
        return type(data)(
            (fill_data_with_source(v, named_data) for v in data)
        )
    elif isinstance(data, NamedDataSource):
        try:
            return named_data.get(data.name)
        except KeyError:
            raise AttributeError(f'Source data "{data.name}" was not found in named data')
    else:
        return data


class NamedTestData:

    def __init__(self, **kwargs):
        self.data = dict(**kwargs)

    @property
    def default(self):
        return list(self.data.items())[0][1]

    def get(self, k):
        """
        Retrieves the object of the given name from the named data, or a sub-property of that object.

        Args:
            k: key of the named data; this could also include reference to a sub-property of the named object
               by using `.` or `__` as the delimiter; examples:
               1. `a.b.c` where `b` is an object - it would try to retrieve data named `a` (main object),
                  then its property `b`, and then finally return `c` as the property of `b`
               2. `a.b.c` where `b` is a mapping - it would try to retrieve data named `a` (main object),
                  then its property `b`, and then finally return `c` as the key of `b`
               3. `a.b.0` where `b` is a sequence (excluding strings) - it would try to retrieve data named
                  `a` (main object), then its property `b`, and then finally return its first value

        Returns:
            retrieved data

        Raises:
            KeyError: if the main object does not exist in the data
            IndexError: if the main object does exist in the data, but the object (or its property) is
                        a sequence that does not have any value at the given index
            AttributeError: if the main object does exist in the data, but the object (or its property):
                            a) is an object and does not have the requested property
                            b) is a mapping and does not have the requested key
        """
        parts = re.split(r'(?:\.|__)', k)

        try:
            value = self.data[parts[0]]
        except KeyError:
            raise KeyError(f'NamedTestData object has no attribute "{parts[0]}"')

        for index, part in enumerate(parts[1:]):
            name = '.'.join(parts[:index + 1])

            try:
                if isinstance(value, Mapping):
                    value = value[part]
                elif isinstance(value, Sequence) and not isinstance(value, str) and part.isnumeric():
                    value = value[int(part)]
                else:
                    value = getattr(value, part)
            except IndexError:
                raise IndexError(f'NamedTestData object "{name}" has no value at index "{part}"')
            except (AttributeError, KeyError):
                raise AttributeError(f'NamedTestData object "{name}" has no attribute "{part}"')

        return value

    def __getattr__(self, k):
        try:
            return self.get(k)
        except (AttributeError, IndexError, KeyError) as e:
            raise AttributeError(str(e)) from e


class NamedDataSource:

    def __init__(self, name):
        self.name = name


def data_provider(function=None, name=None):
    """
    Data provider decorator which prepare data for tests. Data provider should return NamedTestData or name property
    should be set.
    Args:
        function: data provider function or method
        name: name of returned test data

    Returns:
        NamedTestData object
    """

    def _data_provider(function):
        def _decorator(*args, **kwargs):
            return_named_data = kwargs.pop('return_named_data', False)
            returned_data_name = kwargs.pop('returned_data_name', None)
            data = function(*args, **kwargs)
            if return_named_data:
                if not isinstance(data, NamedTestData) and name:
                    data = NamedTestData(**{name: data})
                if returned_data_name:
                    return data.get(returned_data_name)
                else:
                    return data
            else:
                if not isinstance(data, NamedTestData):
                    return data
                elif returned_data_name:
                    return data.get(returned_data_name)
                elif name:
                    return data.get(name)
                else:
                    return data.default
        wrapper = wraps(function)(_decorator)
        wrapper.is_data_provider = True
        return wrapper

    if callable(function):
        return _data_provider(function)
    else:
        return _data_provider


def call(callable, self, data, named_data=None, default_args=None, default_kwargs=None):
    function_or_method_kwargs = {} if default_kwargs is None else dict(**default_kwargs)
    function_or_method_args = [] if default_args is None else list(default_args)

    if 'self' in signature(callable).parameters:
        function_or_method_args.insert(0, self)

    if named_data:
        function_or_method_kwargs.update(
            {
                k: v for k, v in named_data.data.items()
                if k in signature(callable).parameters and k not in function_or_method_kwargs
            }
        )
        return callable(*function_or_method_args, **function_or_method_kwargs)
    else:
        return callable(*function_or_method_args, *data, **function_or_method_kwargs)


def call_test_method(method, self, data, named_data, use_rollback=False):
    if not named_data and isinstance(data, NamedTestData):
        named_data = NamedTestData(**data.data)
    elif named_data and isinstance(data, NamedTestData):
        named_data.data.update(data.data)
    elif named_data and not isinstance(data, NamedTestData):
        named_data = None

    is_data_consumer = getattr(method, 'is_data_consumer', False)
    if use_rollback:
        sid = transaction.savepoint()
    try:
        if is_data_consumer:
            method(self, data=data, named_data=named_data)
        else:
            if hasattr(self, 'set_up_data_consumer'):
                self.tear_up_data_consumer()
            call(method, self, ((data,) if not is_iterable(data) else data), named_data)
    finally:
        if use_rollback:
            transaction.savepoint_rollback(sid)
            responses.reset()
            refresh_model_objects(
                *(data.data.values() if isinstance(data, NamedTestData) else data)
            )
        if not is_data_consumer and hasattr(self, 'tear_down_data_consumer'):
            self.tear_down_data_consumer()


def _copy_named_data(named_data):
    if named_data is None:
        return named_data
    return NamedTestData(**named_data.data)


def _rename_output_data(data, name):
    if not name:
        return data

    if isinstance(name, list):
        assert isinstance(data, (list, tuple, set, NamedTestData)), \
            'Ouput data must be iterable if more output names are set'
        names = name
        values = list(data.data.values()) if isinstance(data, NamedTestData) else list(data)
    else:
        names = [name]
        values = list(data.data.values()) if isinstance(data, NamedTestData) else [data]

    assert len(names) <= len(values), 'More output names than returned values'

    return NamedTestData(**{k: v for k, v in zip(names, values)})


def data_consumer(callable_or_property_or_str, *data_provider_args, **data_provider_kwargs):
    """
    Data provider decorator, allows another callable to provide the data for the test
    Args:
        callable_or_property_or_str: data provider function/method or string name of the data provider method of
                                     the test case class
        *data_provider_args: arguments for the data provider
        **data_provider_kwargs: arguments for the data provider kwargs

    Returns:
        Data created by data provider
    """

    output_name = data_provider_kwargs.pop('_output_name', None)
    def test_decorator(fn):
        def get_data(self, last_data, named_data=None):
            last_args = last_data if last_data else ()
            if isinstance(last_args, NamedTestData):
                last_args = (last_args.default,)
            elif not is_iterable(last_args):
                last_args = (last_args,)

            callable_or_property = (
                getattr(self, callable_or_property_or_str)
                if isinstance(callable_or_property_or_str, str) else callable_or_property_or_str
            )

            if ismethod(callable_or_property) or isclass(callable_or_property) or isfunction(callable_or_property):
                if getattr(callable_or_property, 'is_data_provider', False):
                    data_provider_kwargs['return_named_data'] = True
                data = call(
                    callable_or_property,
                    self,
                    last_args,
                    named_data,
                    fill_data_with_source(data_provider_args, named_data),
                    fill_data_with_source(data_provider_kwargs, named_data)
                )
            else:
                assert isinstance(callable_or_property, (list, tuple, set, types.GeneratorType)), (
                    'Only list, tuple, set, generator, function or method can be used with data_consumer'
                )
                data = callable_or_property

            if isinstance(data, (list, tuple, set, types.GeneratorType)):
                for value in fill_data_with_source(data, named_data):
                    value = _rename_output_data(value, output_name)
                    if not isinstance(value, NamedTestData):
                        value = list(last_args) + (list(value) if isinstance(value, (list, tuple, set)) else [value])
                    yield value, True
            else:
                yield _rename_output_data(data, output_name), False

        def repl(self, data=None, named_data=None):
            for data, use_rollback in get_data(self, data, named_data):
                call_test_method(fn, self, data, _copy_named_data(named_data), use_rollback=use_rollback)

        wrapper = wraps(fn)(repl)
        wrapper.is_data_consumer = True
        wrapper.get_data = get_data
        wrapper.fn = fn
        return wrapper
    return test_decorator


def turn_off_auto_now(model_class, field_name):

    def _turn_off_auto_now(function):
        def _decorator(self, *args, **kwargs):
            field = model_class._meta.get_field(field_name)
            if not isinstance(field, (DateField, DateTimeField)):
                raise RuntimeError('Field %s must be DateField or DateTimeField type') % field_name
            if not field.auto_now:
                raise RuntimeError('Field %s must have set auto_no to True') % field_name
            field.auto_now = False
            try:
                function(self, *args, **kwargs)
            except AssertionError:
                raise
            finally:
                field.auto_now = True
        return wraps(function)(_decorator)

    return _turn_off_auto_now



from django.core.exceptions import ObjectDoesNotExist
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.models.constants import LOOKUP_SEP
from django.test.utils import CaptureQueriesContext
from django.contrib.contenttypes.models import ContentType

from .trivials import assert_equal, assert_true, assert_false


def get_pks(iterable):
    return [obj.pk for obj in iterable]


def assert_iterable_equal(first, second, msg=None):
    assert_equal(set(first), set(second), msg)


def assert_qs_exists(qs, msg=None):
    assert_true(qs.exists(), msg)


def assert_qs_not_exists(qs, msg=None):
    assert_false(qs.exists(), msg)


def assert_qs_contains(qs, obj, msg=None):
    assert_equal(
        set(qs.filter(pk__in=get_pks(obj if isinstance(obj, (set, list, tuple)) else {obj}))),
        set(obj) if isinstance(obj, (set, list, tuple)) else {obj},
        msg
    )


def assert_qs_not_contains(qs, obj, msg=None):
    assert_false(
        qs.filter(pk__in=get_pks(obj if isinstance(obj, (set, list, tuple)) else {obj})).exists(),
        msg
    )


def model_instance_getattr(instance, key):
    try:
        return getattr(instance, key)
    except ObjectDoesNotExist:
        return None


def get_value_from_model_instance(instance, field_name):
    if LOOKUP_SEP in field_name:
        current_field_name, next_field_name = field_name.split(LOOKUP_SEP, 1)
        value = model_instance_getattr(instance, current_field_name)
        if value is None:
            raise AttributeError('Value cannot be get from instance')
        return get_value_from_model_instance(value, next_field_name)
    else:
        return model_instance_getattr(instance, field_name)


def assert_equal_model_fields(instance, refresh_from_db=False, **field_values):
    if refresh_from_db:
        instance.refresh_from_db()
    for field_name, field_value in field_values.items():
        assert_equal(
            get_value_from_model_instance(instance, field_name), field_value, 'Invalid value of "{}"'.format(field_name)
        )


class _AssertNumQueriesContext(CaptureQueriesContext):

    def __init__(self, num, connection, clear_content_type_cache):
        self.num = num
        self.clear_content_type_cache = clear_content_type_cache
        super().__init__(connection)

    def __enter__(self):
        if self.clear_content_type_cache:
            ContentType.objects.clear_cache()
        super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is not None:
            return
        executed = len(self)
        assert_equal(
            executed, self.num,
            "%d queries executed, %d expected\nCaptured queries were:\n%s" % (
                executed, self.num,
                '\n'.join(
                    '%d. %s' % (i, query['sql']) for i, query in enumerate(self.captured_queries, start=1)
                )
            )
        )


def assert_num_queries(num, func=None, *args, using=DEFAULT_DB_ALIAS, clear_content_type_cache=False, **kwargs):
    conn = connections[using]

    context = _AssertNumQueriesContext(num, conn, clear_content_type_cache)

    if func is None:
        return context

    with context:
        func(*args, **kwargs)



from io import StringIO

from contextlib import contextmanager

from django.core.management import call_command
from django.db import DEFAULT_DB_ALIAS, connections


RUN_PRE_COMMIT = "run_pre_commit"
RUN_ON_COMMIT = "run_on_commit"


class CatchCallbacks:

    def __init__(self, connection, callback_name):
        self._connection = connection
        self._callback_name = callback_name
        self._start_count = len(self._watching_callbacks)
        self._end_count = None
        self._executed = False

    @property
    def _watching_callbacks(self):
        return getattr(self._connection, self._callback_name, [])

    def _set_watching_callback(self, callbacks):
        setattr(self._connection, self._callback_name, callbacks)

    def end(self):
        self._end_count = len(self._watching_callbacks)

    def get_callbacks(self, start_count=None):
        start_count = self._start_count if start_count is None else start_count

        if self._end_count is None:
            watching_callbacks = self._watching_callbacks[start_count:]
        else:
            watching_callbacks = self._watching_callbacks[start_count:self._end_count]

        callbacks = []
        for watching_callback in watching_callbacks:
            if self._callback_name == RUN_PRE_COMMIT:
                callbacks.append(watching_callback[-1])
            elif self._callback_name == RUN_ON_COMMIT:
                callbacks.append(watching_callback[-2])
        return callbacks

    def execute(self):
        if self._executed:
            raise RuntimeError('callback was already executed')

        executed_callbacks = []
        start_count = self._start_count
        while True:
            callbacks = self.get_callbacks(start_count)
            if not callbacks:
                break
            executed_callbacks.append(callbacks[0])
            callbacks[0]()
            start_count += 1

        if executed_callbacks:
            self._set_watching_callback(
                self._watching_callbacks[:self._start_count]
                + self._watching_callbacks[self._start_count + len(executed_callbacks):]
            )

        return executed_callbacks


class CommitCallbacks:

    def __init__(self, connection):
        self.pre_commit = CatchCallbacks(connection, RUN_PRE_COMMIT)
        self.on_commit = CatchCallbacks(connection, RUN_ON_COMMIT)

    def end(self):
        self.pre_commit.end()
        self.on_commit.end()

    def execute_pre_commit(self):
        return self.pre_commit.execute()

    def execute_on_commit(self):
        return self.on_commit.execute()

    def execute_on_commit_cascade(self):
        with capture_commit_callbacks(execute_pre_commit=True) as callbacks:
            executed_callbacks = self.execute_on_commit()

        # start count must be updated because current callback removed some values from run_on_commit list
        callbacks.on_commit._start_count -= len(executed_callbacks)
        if callbacks.on_commit.get_callbacks():
            callbacks.execute_on_commit_cascade()


@contextmanager
def capture_commit_callbacks(*, using=DEFAULT_DB_ALIAS, execute_pre_commit=False,
                             execute_on_commit=False, execute_on_commit_cascade=False):
    callbacks = CommitCallbacks(connections[using])
    try:
        yield callbacks
    finally:
        if execute_on_commit_cascade or execute_on_commit or execute_pre_commit:
            callbacks.execute_pre_commit()

        callbacks.end()

        if execute_on_commit_cascade:
            callbacks.execute_on_commit_cascade()
        elif execute_on_commit:
            callbacks.execute_on_commit()


def test_call_command(command, *args,  **kwargs):
    execute_pre_commit = kwargs.pop('execute_pre_commit', True)
    execute_on_commit = kwargs.pop('execute_on_commit', False)
    execute_on_commit_cascade = kwargs.pop('execute_on_commit_cascade', False)
    with capture_commit_callbacks(execute_pre_commit=execute_pre_commit,
                                  execute_on_commit=execute_on_commit,
                                  execute_on_commit_cascade=execute_on_commit_cascade):
        stdout_output = StringIO()
        stderr_output = StringIO()
        call_command(command, stdout=stdout_output, stderr=stderr_output, *args, **kwargs)
    return stdout_output, stderr_output
